import os.path as osp
from collections import OrderedDict
import math
import logging

import torch
import torch.nn as nn
from torch.nn import functional as F
from torch.cuda.amp import GradScaler, autocast

from .dassl_utils import load_checkpoint

from .clip import _MODELS, _download, tokenize
from .clip_model import build_clip_model

from clip.simple_tokenizer import SimpleTokenizer as _Tokenizer

_tokenizer = _Tokenizer()


def load_clip_to_cpu(cfg):
    backbone_name = cfg.MODEL.CLIP_ADAPTER.CLIP_MODEL_NAME
    url = _MODELS[backbone_name]
    model_path = _download(url)

    try:
        # loading JIT archive
        model = torch.jit.load(model_path, map_location="cpu").eval()
        state_dict = None

    except RuntimeError:
        state_dict = torch.load(model_path, map_location="cpu")
    design_details = { "trainer": "VPT",
                    "vision_depth": cfg.MODEL.VPT.PROMPT_DEPTH_VISION,
                      "vision_ctx": cfg.MODEL.VPT.N_CTX_VISION,
                      "language_depth": 0,
                      "language_ctx": 0}
    assert cfg.MODEL.VPT.PROMPT_DEPTH_VISION >= 1, "For Vision Prompting, PROMPT_DEPTH_VISION should be >= 1"
    model = build_clip_model(state_dict or model.state_dict(), design_details)

    return model.float()


class TextEncoder(nn.Module):
    def __init__(self, clip_model):
        super().__init__()
        self.transformer = clip_model.transformer
        self.positional_embedding = clip_model.positional_embedding
        self.ln_final = clip_model.ln_final
        self.text_projection = clip_model.text_projection
        self.dtype = clip_model.dtype

    def forward(self, prompts, tokenized_prompts):
        x = prompts + self.positional_embedding.type(self.dtype)
        x = x.permute(1, 0, 2)  # NLD -> LND
        x = self.transformer(x)
        x = x.permute(1, 0, 2)  # LND -> NLD
        x = self.ln_final(x).type(self.dtype)

        # x.shape = [batch_size, n_ctx, transformer.width]
        # take features from the eot embedding (eot_token is the highest number in each sequence)
        x = x[torch.arange(x.shape[0]), tokenized_prompts.argmax(dim=-1)] @ self.text_projection

        return x


class FixedEmbeddings():
    def __init__(self, cfg, classnames, clip_model):
        clip_imsize = clip_model.visual.input_resolution
        cfg_imsize = cfg.MODEL.VPT.INPUT_SIZE[0]
        assert cfg_imsize == clip_imsize, f"cfg_imsize ({cfg_imsize}) must equal to clip_imsize ({clip_imsize})"

        prompt_prefix = "a photo of a"
        logging.info('Vision Prompting Design')
        logging.info(f'Initial context: "{prompt_prefix}"')
        logging.info(f"Number of context words (tokens) for Vision prompting: {cfg.MODEL.VPT.N_CTX_VISION}")
        logging.info(f"Using fixed hand crated prompts")

        classnames = [name.replace("_", " ") for name in classnames]
        prompts = [prompt_prefix + " " + name + "." for name in classnames]

        tokenized_prompts = torch.cat([tokenize(p) for p in prompts])
        with torch.no_grad():
            text_features = clip_model.encode_text(tokenized_prompts)

        self.fixed_embeddings = text_features

    def return_fixed_embeddings(self):
        return self.fixed_embeddings


class CustomCLIP(nn.Module):
    def __init__(self, cfg, classnames, clip_model):
        super().__init__()
        self.embeddings = FixedEmbeddings(cfg, classnames, clip_model)
        self.image_encoder = clip_model.visual
        self.text_encoder = TextEncoder(clip_model)
        self.logit_scale = clip_model.logit_scale
        self.dtype = clip_model.dtype

    def forward(self, image, label=None, training=False):
        logit_scale = self.logit_scale.exp()

        text_features = self.embeddings.return_fixed_embeddings().cuda()
        image_features = self.image_encoder(image.type(self.dtype))

        image_features = image_features / image_features.norm(dim=-1, keepdim=True)
        text_features = text_features / text_features.norm(dim=-1, keepdim=True)
        logits = logit_scale * image_features @ text_features.t()

        if training:
            return F.cross_entropy(logits, label)

        return logits
    
    def get_image_features(self, image):
        image_features = self.image_encoder(image.type(self.dtype)).float()
        image_features = image_features / image_features.norm(dim=-1, keepdim=True)
        return image_features


def build_model(cfg, class_names, clip_model):
    logging.info("Building CustomCLIP_VPT")
    model = CustomCLIP(cfg, class_names, clip_model)

    logging.info("Turning off all gradients")
    for name, param in model.named_parameters():
        param.requires_grad_(False)

    # Double check
    enabled = set()
    for name, param in model.named_parameters():
        if param.requires_grad:
            enabled.add(name)
    logging.info(f"Parameters to be updated: {enabled}")

    # Count the number of parameters to be updated
    num_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    logging.info(f"Number of parameters to be updated: {num_params}")

    # if cfg.MODEL.INIT_WEIGHTS:
    #     load_pretrained_weights(self.model, cfg.MODEL.INIT_WEIGHTS)

    # model.to(self.device)
    # NOTE: only give prompt_learner to the optimizer
    # self.register_model("prompt_learner", self.model, self.optim, self.sched)

    # Note that multi-gpu training could be slow because CLIP's size is
    # big, which slows down the copy operation in DataParallel
    device_count = torch.cuda.device_count()
    if device_count > 1:
        logging.info(f"Multiple GPUs detected (n_gpus={device_count}), use all of them!")
        # model = nn.DataParallel(model)
        raise NotImplementedError

    return model

def load_model(model, cfg):
    directory = cfg.MODEL.VPT.DIR
    epoch = cfg.MODEL.VPT.LOAD_EPOCH
    if not directory:
        logging.info("Note that load_model() is skipped as no pretrained model is given")
        return

    # By default, the best model is loaded
    model_file = "model-best.pth.tar"

    if epoch is not None:
        model_file = "model.pth.tar-" + str(epoch)

    for name in ["prompt_learner"]:
        model_path = osp.join(directory, name, model_file)

        if not osp.exists(model_path):
            raise FileNotFoundError('Model not found at "{}"'.format(model_path))

        checkpoint = load_checkpoint(model_path)
        state_dict = checkpoint["state_dict"]
        epoch = checkpoint["epoch"]

        # Ignore fixed token vectors
        if "prompt_learner.token_prefix" in state_dict:
            del state_dict["prompt_learner.token_prefix"]

        if "prompt_learner.token_suffix" in state_dict:
            del state_dict["prompt_learner.token_suffix"]

        logging.info("Loading weights to {} " 'from "{}" (epoch = {})'.format(name, model_path, epoch))
        model.load_state_dict(state_dict, strict=False)
    
    return model
